// z64.h: 2015-05-25 sieve of Eratosthenes for prime numbers, ver.4
// (c) Khashin S.I.  khash2@gmail.com

typedef unsigned long long u64;
typedef          long long i64;

static const u64     MAX64=18446744073709551615ULL;
static const u64     MAX32=4294967295ULL;

class z64 {
private:
//public:
	static unsigned nMax32;			// sieve-32 array size, even
	static unsigned N032;			// =nMax/2
	static unsigned NPrimes32;		// number of primes-32 (including 2)
	static int      *pr32;			// sieve array pr[0..N0-1]
									// if m=2*k+1 prime, then pr[k]= pi[m],
									// else                   pr[k]=-(minimal prime divisor of m)
	static unsigned *pi32;			// array of primes


public:
	static char path[256];			// path to sieve

//private:
public:
	// sieve-32 public functions
	static void     init32(unsigned N=1024*256);	// create & calc array 
private:
	static unsigned maxP32(){return pi32[NPrimes32-1];}//return maximal prime in the sieve
	static unsigned ip32( unsigned n);				// index of prime number, 0 if n is composite
	static unsigned primeFactor32(unsigned n);		// prime divisor or 1, if not found
	static unsigned NextPrime32(unsigned n);		// next prime
	static unsigned PrevPrime32(unsigned n);		// previous prime

private:
	static unsigned size;			// sieve array size
	static unsigned char *v;		// sieve array v[0..size-1]
	static u64 maxPrime;			// maximal prime number in sieve
	static u64 NPrimes;				// total number of primes (including 2)
	static unsigned nBlocks;		// number of blocks in sieve
	static unsigned *w;				// Pi(1024*i)

	static void SetComposite(u64 n);// set n composite
	static void SetComposit2(u64 n);// set composite 3*n, 5*n,...
	static void fillFinish();		// finish fill (maxPrime, NPrimes, v8, ...)

	// sieve of Eratosthenes private functions
	static unsigned bsearchP( unsigned k);	// return  i: w[i] <= k <= w[i+1]
	static bool		EisPrime( u64 n);		// is prime(n)?
	static u64		EithPrime( u64 i64);	// i-th prime number,  Maple: ithprime(i+1)
	//
	static u64 factor_Pollard_rho (u64 n);// Pollard \rho -method of number factorization

	// for PrevPrime, NextPrime
	static int n33;						// the number of division tests
	static unsigned char *p33, *r33;	// primes and remainders
	static void fill33(u64 n);			// fill p32[], r32[]
	static void inc33( int c );			// add r32[]+=c mod p32[]
	static bool isZero33();				// true, if one of r32[] ==0

public:
	// constructor and destructor do nothing
	z64 (){};
	~z64(){};

	static void    toString(char *s, u64 n, int w=0);		// s <- to String of width w
	static char   *toString(u64 n, int w=0);				// return to String of width w
	// 64-bit calculations:
	static bool	   isFullSq (u64 n);						// n is full square?
	static u64	   gcd      (u64 a, u64 b);					// GCD(a,b)
	static u64     mul_mod  (u64 a, u64 b, u64 n);			// (a*b)%c
	static u64     pow_mod  (u64 a, u64 b, u64 c);			// (a^b)%c
	static bool    MillerRabin(u64 a,u64 n);					// MR primality test for n with base a
	static void    euclid   (u64 a, u64 b, i64 &x, i64 &y, u64 &d); // find x,y,d: a*x+b*y=d=gcd(a,b)
	static u64     inverse  (u64 a, u64 n);					// 1/a mod n
	static bool    chineese (u64 mod1, u64 b1, u64 mod2, u64 b2, u64 &mod3, u64 &b3); // cheneese theorem

	// sieve of Eratosthenes
	static void   Efill(u64 mN=16*1024*1024);	// fill sieve for primes till mN
	static void   Efree();					// free memory
	static int    Esave(const char *fname);	// return error code
	static int    Eload(const char *fname, u64 mx=0);	// return error code
	// if( z64::Eload("C:\\a\\primes\\sieve16M.dat") ) return; // till 2^24
	// if( z64::Eload("C:\\a\\primes\\sieve.dat"   ) ) return; // till 2^32
	// if( z64::Eload("C:\\a\\primes\\sieve30_30G.dat"), 1<<30 ) return; // till 2^30

	//
	static u64 maxN();				// return maximal N
	static u64 maxP();				// return maximal prime in the sieve

	// Primes
	static bool isPrime( u64 n);	// is prime(n)?
	static u64 p ( u64 i);			// i-th prime,  Maple: ithprime(i), p(0)=0, p(1)=2, p(2)=3,...
	static u64 ip( u64 n);			// index of prime number, 0 if n is composite
	static u64 Pi( u64 n);			// the number of primes <= n, Pi(0)=Pi(1)=0, Pi(2)=1, Pi(3)=Pi(4)=2,...
	static u64 NextPrime( u64 n);	// prime p>n, NextPrime(3)=5, NextPrieme(4)=5, NextPrime(0)=NextPrime(1)=2, NextPrime(2)=3
	static u64 PrevPrime( u64 n);	// prime p<n, PrevPrime(0)=PrevPrime(1)=PrevPrime(2)=0, PrevPrime(4)=3

	// Factoring
	static u64  primeFactor( u64 n);	// prime divisor or 1, if not found

	// Rest
	static u64  order( u64 a, u64 n);	// order of a by modulo prime n
	static int  jacobi(u64 a, u64 n);	// Jacobi symbol
	static char *factorString(u64 n);	// string with factorisation: 180200 -> 2^3*5^2*17*53
	static u64  msqrt( u64 a, u64 p);	// sqrt(a) mod p, p must be prime!

	// Frobenius
	static void   Frobenius( u64 a, u64 b, i64 c, u64 k, u64 p, u64 &a1, u64 &b1);// a1+b1*sqrt(c) = (a+b*sqrt(c))^k mod p
	static u64    FrobOrder( u64 a, u64 b, i64 c, u64 p);	// order(a+b*sqrt(c)) mod p
	static bool   FrobTest ( u64 a, u64 b, i64 c, u64 n);	// n is prime?
	static int    minC(u64 n);								// minimal c: J(c/n)=-1;
};
//
void Eload(u64 maxN);	//        disk:\a\primes
